// *****************************************************************************************************************
// Imaging Toolbox Script Copyright (C) Blackwater Skies (www.blackwaterskies.co.uk) 2013-16, all rights reserved. *
// *****************************************************************************************************************
//
// V1.2
//   - Had to revert use of https back to http due to CDS not supporting it suddenly!
//   - Fixed create share link to use correct Wordpress permalink instead of relying on redirection rule.
//
//   - Mixed use of http and https in Wordpress requires the following in /etc/apache2/sites-available/blackwaterskies.co.uk.conf :
//        # Rewrite rules: See also blackwaterskies-le-ssl.co.uk.conf
//        # Note that permalinks on in Wordpress also uses mod_rewrite
//        # so that any URIs not matching a file or directory are rewritten
//        # to index.php so that it can redirect the URI to the correct file/directory.
//        # Therefore any rewriting must avoid matching on /index.php
//        RewriteEngine on
//
//        # Redirect to canonical hostname
//        RewriteCond %{HTTP_HOST} !^www\.blackwaterskies\.co\.uk [NC]
//        RewriteCond %{HTTP_HOST} !^$ [NC]
//        RewriteRule ^/(.*) http://www.blackwaterskies.co.uk/$1 [NC,NE,R=permanent,L]
//
//        # Redirect to http for imaging toolbox pages, to https for everything else
//        RewriteCond %{HTTPS} =off
//        RewriteCond %{REQUEST_URI} !^/((imaging-toolbox|index.php)(.*)|wp-content/themes/canvas-child.bws/(.*)\.json) [NC]
//        RewriteRule ^/?(.*) https://www.blackwaterskies.co.uk/$1 [NC,NE,R=permanent,L]
//
//   - Mixed use of http and https in Wordpress requires the following in /etc/apache2/sites-available/blackwaterskies.co.uk-le-ssl.conf :
//        # Rewrite rules: See also blackwaterskies.co.uk.conf
//        # Note that permalinks on in Wordpress also uses mod_rewrite
//        # so that any URIs not matching a file or directory are rewritten
//        # to index.php so that it can redirect the URI to the correct file/directory.
//        # Therefore any rewriting must avoid matching on /index.php
//        RewriteEngine on
//
//        # Force redirect to canonical hostname
//        RewriteCond %{HTTP_HOST} !^www\.blackwaterskies\.co\.uk [NC]
//        RewriteCond %{HTTP_HOST} !^$ [NC]
//        RewriteRule ^/(.*) https://www.blackwaterskies.co.uk/$1 [NC,NE,R=permanent,L]
//
//        # Redirect to http for imaging toolbox pages, to https for everything else
//        RewriteCond %{HTTPS} =on
//        RewriteCond %{REQUEST_URI} ^/imaging-toolbox(.*) [NC]
//        RewriteRule ^/?(.*) http://www.blackwaterskies.co.uk/$1 [NC,NE,R=permanent,L]
//
//        RewriteCond %{HTTPS} =on
//        RewriteCond %{REQUEST_URI} ^/wp-content/themes/canvas-child-bws/(.*)\.json [NC]
//        RewriteRule ^/?(.*) http://www.blackwaterskies.co.uk/wp-content/themes/canvas-child-bws/$1.json [NC,NE,R=permanent,L]
//
// V1.1 (Planned)
// FIXME: Add feature if poss to see if target list can be exported to SGPro.
// FIXME: Add feature select colourmap / reverse.
// FIXME: Add feature to use Aladin functions for surveys around current location.
// FIXME: On touch devices, location display shows NAN during/after dragging survey location.
// FIXME: Aladin Lite bug - constellation sources get 'stuck' when zoomed right out.
// FIXME: On Chrome, Constellation and RA/Dec Lines not drawing correct thickness, colour or at all in some cases. Works fine on Edge.
//
// V1.1 (Current)
//   - Migrated to https, fixed problems with insecure script loading.
//   - Fixed export to png as was not working due to Chrome not allowing top level data urls
//
// V1.0
//   - Wordpress Version
//   - New responsive layout and touch-friendly UI.
//   - New colour picker for reticules.
//   - Split Style select in to Maximise and Notes slider checkboxes.
//   - Constellation lines and names display
//   - Equatorial grid display
//   - Display Messier / Caldwell objects
//
// V0.8
//   - Changed hosting to Github Pages
//
// V0.7
//   - New in Aladin Lite:
//     - Now using Aladin Lite V2.
//     - Updated available surveys:
//       - Default survey is now main version of DSS Colour; any existing links referencing a removed survey will default to this.
//       - Backup DSS survey removed from options for simplicity.
//       - SDSS7 survey replaced with SDSS9 survey.
//       - New VTSS-Ha survey added.
//     - Better control of zooming/locking the field of view:
//       - Eliminated "bouncing" at maximum FOV.
//       - Elimated ineffcient redrawing to force a particular FOV when reticle maximised.
//     - Export view to PNG.
//   - New features:
//     - Split in to separate HTML, CSS and JS files for performance and ease of maintenance:
//       - Can now add/change cameras and scopes without having to do a version release.
//     - Using arctan instead of small angle approximation to calculate reticule sizes.
//       - More accurate for big chips/short focal lengths, thus may result in *slightly* different results.
//     - The 'Link' box under 'Share and Save' automatically selects the whole URL text when clicked.
//
// V0.6
//   - New in Aladin Lite:
//     - Zoom in/out buttons.
//   - New Features:
//     - Support for tablets/smartphones, Aladin Lite display re-sizes to fit.
//       - Redesigned forms, better browser compatibility and flexible layout for mobile devices.
//       - Removed 'Simple Forms' option as hopefully no longer required.
//     - Create and name multiple reticules and compare them on screen simultaneously.
//       - Label each reticule with a description.
//       - Save/load current set of reticules.
//     - Mosaic mode:
//       - Display and offset multiple reticules to plan a mosaic.
//     - Refactored code, more object orientation, doubtless new bugs to find.
//   - Bug Fixes:
//     - Limit Aladin Lite zoom to 90 degrees when reticule displayed to minimise scaling/projection issues.
//   - Known Issues:
//     - The Aladin Lite Sinus projection is used by default.  This can result in some distortion of the survey background
//       when zooming at different scales, especially close to the celestial poles.
//
// V0.5
//   - Initial Release

// ********** Initialise globals and document *******************************************************************

// Global array to hold FOV and form objects, plus track current reticule.
var fovMain = [],
    formMain = [],
    curRet = 0;

// Globals to hold Aladin Lite, catalogs and overlays.
var aladin,
    constellationLinesOverlay,
    constellationNamesCatalog,
    constellationsVisible = false,
    equatorOverlay,
    equatorCatalog,
    cardinalOverlay,
    cardinalCatalog,
    raLinesOverlay,
    raLinesCatalog,
    decLinesOverlay,
    decLinesCatalog,
    equatorialGridVisible = false,
    caldwellCatalog,
    caldwellVisible = false,
    messierCatalog,
    messierVisible = false;

// DEBUG: Global to hold testing overlay.
// var testOverlay;

// Modify elements and initialise Aladin Lite once DOM ready.
$(document).ready(
  function() {
    // Load and parse the targets from JSON formatted file (array of objects) and populate the searchPopular select.
    $.getJSON("/wp-content/themes/canvas-child-bws/targets.json", function(targetSet) {
      for (var i = 0, len = targetSet.length; i < len; i++) {
        // Add target to select.
        if (targetSet[i].targetSelected) {
          $('#searchPopular').append($('<option></option>')
            .prop('value', targetSet[i].targetVal)
            .text(targetSet[i].targetText)
            .prop('selected', true));
        } else {
          $('#searchPopular').append($('<option></option>')
            .prop('value', targetSet[i].targetVal)
            .text(targetSet[i].targetText));
        }
      }
    });

    // Load and parse the cameras from JSON formatted file (array of objects) and populate the equipCamera select.
    $.getJSON("/wp-content/themes/canvas-child-bws/cameras.json", function(cameraSet) {
      for (var i = 0, len = cameraSet.length; i < len; i++) {
        // Add camera to select.
        if (cameraSet[i].cameraSelected) {
          $('#equipCamera').append($('<option></option>')
            .prop('value', cameraSet[i].cameraVal)
            .text(cameraSet[i].cameraText)
            .prop('selected', true));
        } else {
          $('#equipCamera').append($('<option></option>')
            .prop('value', cameraSet[i].cameraVal)
            .text(cameraSet[i].cameraText));
        }
      }
    });

    // Load and parse the scopes from JSON formatted file (array of objects) and populate the equipScope select.
    $.getJSON("/wp-content/themes/canvas-child-bws/scopes.json", function(scopeSet) {
      for (var i = 0, len = scopeSet.length; i < len; i++) {
        // Add scope to select.
        if (scopeSet[i].scopeSelected) {
          $('#equipScope').append($('<option></option>')
            .prop('value', scopeSet[i].scopeVal)
            .text(scopeSet[i].scopeText)
            .prop('selected', true));
        } else {
          $('#equipScope').append($('<option></option>')
            .prop('value', scopeSet[i].scopeVal)
            .text(scopeSet[i].scopeText));
        }
      }
    });

    // Initialise Aladin Lite.
    // Key name               Default value         Description
    // target                 0 +0                  Initial target, as a position or an object name resolved by Sesame
    // cooFrame               "ICRS"                Coordinate system: "ICRS" or "galactic" *** NB Probably incorrect, use "j2000" instead of "ICRS"
    // survey                 "P/DSS/color"         Identifier of the initial image survey.
    // fov                    60                    Initial value of the visible field of view, in decimal degrees
    // showReticle            true                  If true, the reticle will be displayed
    // showZoomControl        true                  If true, the zoom control GUI is displayed (plus/minus buttons)
    // showFullscreenControl  true                  If true, the button to pass in full screen mode (at the top right of the interface) is displayed
    // showLayersControl      true                  If true, the icon to open the layers controls is displayed
    // showGotoControl        true                  If true, the icon to easily jump to a new position/object is displayed
    // showShareControl       false                 If true, show the share control
    // showCatalog            true                  If true, show catalog
    // showFrame              true                  If true, coordinate are displayed (at the top left of the interface)
    // fullScreen             false                 If true, Aladin Lite starts in "full screen" mode
    // reticleColor           "rgb(178, 50, 178)"   Color of the reticle
    // reticleSize            22                    Size in pixels of the reticle
    // log                    true                  If true, enable logging (to where?)
    // allowFullZoomout       false                 If true, allow zoom out to full extent?
    aladin = A.aladin('#aladin-lite-div', {
      target: "Horsehead Nebula",
      cooFrame: "j2000",
      survey: "P/DSS2/color",
      fov: 2.19,
      showReticle: true,
      showZoomControl: true,
      showFullscreenControl: true,
      showLayersControl: false,
      showGotoControl: false,
      showShareControl: false,
      showCatalog: true,
      showFrame: false,
      fullScreen: false,
      reticleColor: "rgb(0, 255, 0)",
      reticleSize: 22,
      log: false,
      allowFullZoomout: true
    });

    // Load and parse the constellation outlines.
    $.getJSON("/wp-content/themes/canvas-child-bws/constellationlines.json", function(constellationLines) {
      constellationLinesOverlay = A.graphicOverlay({color: 'white', lineWidth: 1});
      aladin.addOverlay(constellationLinesOverlay);
      constellationLinesOverlay.hide();
      for (var i = 0, len = constellationLines.length; i < len; i++) {
        // Add Aladin polyline objects from JSON to constellation overlay.
        constellationLinesOverlay.add(A.polyline(constellationLines[i]));
      }
    });

    // Load and parse the constellation names.
    $.getJSON("/wp-content/themes/canvas-child-bws/constellationnames.json", function(constellationNames) {
      constellationNamesCatalog = A.catalog({name: 'constellationNamesCatalog', sourceSize: 0, color: "rgba(0,0,0,0)", displayLabel: true, labelColumn: 'sourceLabel', labelColor: 'white', labelFont: '400 20px Hind'});
      aladin.addCatalog(constellationNamesCatalog);
      constellationNamesCatalog.hide();
      for (i = 0, len = constellationNames.length; i < len; i++) {
        // Add catalog source with constellation name.
        constellationNamesCatalog.addSources(A.source(constellationNames[i][1], constellationNames[i][2], {sourceLabel: constellationNames[i][0]}));
      }
    });

    // Load and parse the equatorial grid.
    $.getJSON("/wp-content/themes/canvas-child-bws/equator.json", function(equator) {
      equatorOverlay = A.graphicOverlay({color: 'green', lineWidth: 1});
      aladin.addOverlay(equatorOverlay);
      equatorOverlay.hide();
      for (var i = 0, len = equator.length; i < len; i++) {
        // Add Aladin polyline objects from JSON to equator overlay.
        equatorOverlay.add(A.polyline(equator[i]));
      }
    });

    $.getJSON("/wp-content/themes/canvas-child-bws/equatorcatalog.json", function(equatorC) {
      equatorCatalog = A.catalog({name: 'equatorCatalog', sourceSize: 0, color: "rgba(0,0,0,0)", displayLabel: true, labelColumn: 'sourceLabel', labelColor: 'green', labelFont: '300 16px Hind'});
      aladin.addCatalog(equatorCatalog);
      equatorCatalog.hide();
      for (i = 0, len = equatorC.length; i < len; i++) {
        // Add catalog source with dec coordinates.
        equatorCatalog.addSources(A.source(equatorC[i][1], equatorC[i][2], {sourceLabel: equatorC[i][0]}));
      }
    });

    $.getJSON("/wp-content/themes/canvas-child-bws/cardinal.json", function(cardinal) {
      cardinalOverlay = A.graphicOverlay({color: 'gold', lineWidth: 1});
      aladin.addOverlay(cardinalOverlay);
      cardinalOverlay.hide();
      for (var i = 0, len = cardinal.length; i < len; i++) {
        // Add Aladin polyline objects from JSON to cardinal overlay.
        cardinalOverlay.add(A.polyline(cardinal[i]));
      }
    });

    $.getJSON("/wp-content/themes/canvas-child-bws/cardinalcatalog.json", function(cardinalC) {
      cardinalCatalog = A.catalog({name: 'cardinalCatalog', sourceSize: 0, color: "rgba(0,0,0,0)", displayLabel: true, labelColumn: 'sourceLabel', labelColor: 'gold', labelFont: '300 16px Hind'});
      aladin.addCatalog(cardinalCatalog);
      cardinalCatalog.hide();
      for (i = 0, len = cardinalC.length; i < len; i++) {
        // Add catalog source with RA coordinates.
        cardinalCatalog.addSources(A.source(cardinalC[i][1], cardinalC[i][2], {sourceLabel: cardinalC[i][0]}));
      }
    });

    $.getJSON("/wp-content/themes/canvas-child-bws/ralines.json", function(raLines) {
      raLinesOverlay = A.graphicOverlay({color: 'gray', lineWidth: 1});
      aladin.addOverlay(raLinesOverlay);
      raLinesOverlay.hide();
      for (var i = 0, len = raLines.length; i < len; i++) {
        // Add Aladin polyline objects from JSON to RA lines overlay.
        raLinesOverlay.add(A.polyline(raLines[i]));
      }
    });

    $.getJSON("/wp-content/themes/canvas-child-bws/ralinescatalog.json", function(raLinesC) {
      raLinesCatalog = A.catalog({name: 'raLinesCatalog', sourceSize: 0, color: "rgba(0,0,0,0)", displayLabel: true, labelColumn: 'sourceLabel', labelColor: 'gray)', labelFont: '300 16px Hind'});
      aladin.addCatalog(raLinesCatalog);
      raLinesCatalog.hide();
      for (i = 0, len = raLinesC.length; i < len; i++) {
        // Add catalog source with RA coordinates.
        raLinesCatalog.addSources(A.source(raLinesC[i][1], raLinesC[i][2], {sourceLabel: raLinesC[i][0]}));
      }
    });

    $.getJSON("/wp-content/themes/canvas-child-bws/declines.json", function(decLines) {
      decLinesOverlay = A.graphicOverlay({color: 'gray', lineWidth: 1});
      aladin.addOverlay(decLinesOverlay);
      decLinesOverlay.hide();
      for (var i = 0, len = decLines.length; i < len; i++) {
        // Add Aladin polyline objects from JSON to RA lines overlay.
        decLinesOverlay.add(A.polyline(decLines[i]));
      }
    });

    $.getJSON("/wp-content/themes/canvas-child-bws/declinescatalog.json", function(decLinesC) {
      decLinesCatalog = A.catalog({name: 'decLinesCatalog', sourceSize: 0, color: "rgba(0,0,0,0)", displayLabel: true, labelColumn: 'sourceLabel', labelColor: 'gray', labelFont: '300 16px Hind'});
      aladin.addCatalog(decLinesCatalog);
      decLinesCatalog.hide();
      for (i = 0, len = decLinesC.length; i < len; i++) {
        // Add catalog source with dec coordinates.
        decLinesCatalog.addSources(A.source(decLinesC[i][1], decLinesC[i][2], {sourceLabel: decLinesC[i][0]}));
      }
    });

    // Load and parse the Caldwell catalog.
    $.getJSON("/wp-content/themes/canvas-child-bws/caldwell.json", function(caldwell) {
      caldwellCatalog = A.catalog({name: 'caldwellCatalog', sourceSize: 10, shape: "cross", color: "rgb(0,255,195)", displayLabel: true, labelColumn: 'sourceLabel', labelColor: 'rgb(0,255,195)', labelFont: '400 16px Hind'});
      aladin.addCatalog(caldwellCatalog);
      caldwellCatalog.hide();
      for (i = 0, len = caldwell.length; i < len; i++) {
        // Add catalog source with Caldwell number.
        caldwellCatalog.addSources(A.source(caldwell[i][1], caldwell[i][2], {sourceLabel: caldwell[i][0]}));
      }
    });

    // Load and parse the Messier catalog.
    $.getJSON("/wp-content/themes/canvas-child-bws/messier.json", function(messier) {
      messierCatalog = A.catalog({name: 'messierCatalog', sourceSize: 10, shape: "plus", color: "rgb(255,60,0)", displayLabel: true, labelColumn: 'sourceLabel', labelColor: 'rgb(255, 60, 0)', labelFont: '400 16px Hind'});
      aladin.addCatalog(messierCatalog);
      messierCatalog.hide();
      for (i = 0, len = messier.length; i < len; i++) {
        // Add catalog source with Messier number.
        messierCatalog.addSources(A.source(messier[i][1], messier[i][2], {sourceLabel: messier[i][0]}));
      }
    });

    // Add the FOV Locked indicator element and hide it.
    $('.aladin-fov').after('<div class="bws-maxIndicator"> (Locked)</div>');
    $('.bws-maxIndicator').css('display', 'none');
  }
);

// Initialise everything else once window loaded.
$(window).load(function() {
  var gotReticules = false, // Flag whether we got any reticules previously stored.
    finalZoom = 2.19,       // Default final zoom.
    i;                      // Loop counter.

  // Check local storage for a set of saved reticules.
  if (localStorage.getItem('reticuleSet') !== null) {
    var reticuleSet = JSON.parse(localStorage.getItem('reticuleSet'));
    gotReticules = true;

    // Extract the set of reticules.
    for (i = 0, len = reticuleSet.length; i < len; i++) {
      fovMain[i] = new fovReticule(2.13, 1.42, 0, 0, 0, '#00FF00', 0, 0, 0, true, true, 'New');
      formMain[i] = new itForm();
      aladin.addCanvas(fovMain[i].guid);
      formMain[i].retrieveSavedForm(reticuleSet[i]);
    }

    // Make the last viewed reticule current.
    if (localStorage.getItem('curRet') !== null) {
      curRet = localStorage.getItem("curRet") * 1;
    }

    // Record desired final zoom.
    finalZoom = reticuleSet[curRet].zoom;
  } else {
    // Create a default reticule and form.
    fovMain[curRet] = new fovReticule(2.13, 1.42, 0, 0, 0, '#00FF00', 0, 0, 0, true, true, 'New');
    formMain[curRet] = new itForm();
    aladin.addCanvas(fovMain[curRet].guid);
  }

  // Check for saved settings in URL.
  var savedUrl = window.getUrlParam('BWSFOV');
  if (savedUrl) {
    if (gotReticules) {
      // Create a default reticule and form to avoid overwriting those retrieved from local storage.
      curRet = fovMain.length;
      fovMain[curRet] = new fovReticule(2.13, 1.42, 0, 0, 0, '#00FF00', 0, 0, 0, true, true, 'New');
      formMain[curRet] = new itForm();
      aladin.addCanvas(fovMain[curRet].guid);
    } else {
      gotReticules = true;
    }

    // Extract the saved settings from the URL and put them in to the UI.
    var saveForm = JSON.parse(decodeURIComponent(savedUrl));
    formMain[curRet].retrieveSavedForm(saveForm);

    // Update desired final zoom.
    finalZoom = saveForm.zoom;
  }

  // Initialise UI handling.
  initUI();

  // If we retrieved any reticules then display them.
  if (gotReticules) {
    for (i = 0, len = formMain.length; i < len; i++) {
      if (formMain[i].compDisplay) {
        if (i != curRet) {
          // This is a background reticule.
          fovMain[i].toBack();
        } else {
          fovMain[i].toFore();
        }

        // Populate the FOV reticule with retrieved values.
        fovMain[i].wDeg = formMain[i].fovWidth;
        fovMain[i].hDeg = formMain[i].fovHeight;
        fovMain[i].rDeg = formMain[i].fovRotation;
        fovMain[i].hOff = formMain[i].fovYOffset;
        fovMain[i].wOff = formMain[i].fovXOffset;
        fovMain[i].maxOvl = formMain[i].fovMax;
        fovMain[i].annotateOvl = formMain[i].fovNotes;
        fovMain[i].drawColour = formMain[i].fovColour;
        fovMain[i].appH = formMain[i].fovAppH;
        fovMain[i].appV = formMain[i].fovAppV;
        fovMain[i].scopeRes = formMain[i].fovScopeRes;
        fovMain[i].description = formMain[i].compDesc;

        // Display the FOV reticule.
        fovMain[i].showFov(aladin);
      }
    }

    // Update the form UI.
    formMain[curRet].putAllForm();
    if (formMain[curRet].compDisplay) {
      formMain[curRet].buttonFovForm(String.fromCodePoint(0xE8F4));
    }

    // Go to the retrieved RA/Dec.
    $('#gotoRaDec').click();

    // Zoom to the desired FOV.
    aladin.progressiveFov(finalZoom, false);

    // Display the required survey.
    aladin.setImageSurvey(formMain[curRet].searchSurveySelection);
  }

  // Update the form UI some more.
  formMain[curRet].cleanUrlForm();

  // Update the Compare form UI to match the set of reticules loaded/created.
  compNavigate();
});

// ********** Functions *****************************************************************************************

// *** Extend Aladin Lite to add a canvas for drawing the FOV reticule.
if (typeof(Aladin.prototype.addCanvas) === 'undefined') {
  Aladin.prototype.addCanvas = function(canvId) {
    // Create additional canvas for FoV overlay.
    var fovCanvas = document.createElement('canvas');

    fovCanvas.id = canvId;
    fovCanvas.width = aladin.getSize()[0];
    fovCanvas.height = aladin.getSize()[1];
    fovCanvas.style.zIndex = 2;
    fovCanvas.style.position = 'absolute';
    fovCanvas.style.left = 0;
    fovCanvas.style.top = 0;

    // Append FOV canvas to the aladin lite div.
    $('#' + this.aladinDiv.id).append(fovCanvas);

    if (!this.hasOwnProperty('ctx')) {
      this.ctx = [];
    }

    // Get context of new canvas.
    this.ctx[canvId] = $('#' + fovCanvas.id)[0].getContext('2d');

    // Set origin of canvas to centre of view.
    this.ctx[canvId].translate(this.getSize()[0] / 2, this.getSize()[1] / 2);
  };
}

// *** Extend Aladin Lite to resize a previously added canvas to match current size of Aladin view.
if (typeof(Aladin.prototype.resizeCanvas) === 'undefined') {
  Aladin.prototype.resizeCanvas = function(canvId) {
    // Reset the transform so origin is in default positon.
    this.ctx[canvId].setTransform(1, 0, 0, 1, 0, 0);

    // Resize canvas.
    this.ctx[canvId].canvas.width = this.getSize()[0];
    this.ctx[canvId].canvas.height = this.getSize()[1];

    // Set origin for canvas to centre of view.
    this.ctx[canvId].translate(this.getSize()[0] / 2, this.getSize()[1] / 2);
  };
}

// *** Extend Aladin Lite to remove a previously added canvas.
if (typeof(Aladin.prototype.removeCanvas) === 'undefined') {
  Aladin.prototype.removeCanvas = function(canvId) {
    $('#' + canvId).remove();
    delete this.ctx[canvId];
  };
}

// *** Extend Aladin Lite to progressively change FOV until >= to requiredFov.
if (typeof(Aladin.prototype.progressiveFov) === 'undefined') {
  Aladin.prototype.progressiveFov = function(requiredFov, smaller) {
    // Aladin Lite will only set FOV in increments depending on the projection and frame size.
    // This method ensures that we always end up with a FOV that is at least as wide as | no wider than the one we request.
    // Request the required FOV
    this.setFov(requiredFov);

    // Increase the FOV in single steps until it is...
    // (smaller == false) greater than or equal to the FOV requested.
    // (smaller == true) less than or equal to the FOV requested.
    for (;(smaller) ? this.getFov()[0] > requiredFov : this.getFov()[0] < requiredFov;) {
      this.decreaseZoom(1);
    }
  };
}

// *** Create animation methods.
if (typeof(window.fovReqAnim) === 'undefined') {
  var lastTime = 0,
      vendorsR = ['r', 'msR', 'mozR', 'webkitR', 'oR'],
      vendorsC = ['c', 'msC', 'mozC', 'webkitC', 'oC'];

  // Loop through vendors to try to use supported RequestAnimationFrame and CancelAnimationFrame.
  for (var x = 0; x < vendorsR.length && typeof(window.fovReqAnim) === "undefined"; ++x) {
    window.fovReqAnim = window[vendorsR[x] + 'equestAnimationFrame'];
    window.fovCanAnim = window[vendorsC[x] + 'ancelAnimationFrame'] ||
      window[vendorsC[x] + 'ancelRequestAnimationFrame'];
  }

  if (typeof(window.fovReqAnim) === 'undefined' || typeof(window.fovCanAnim) === 'undefined') {
    // Fail back to using a setTimeout / clearTimeout.
    window.fovReqAnim = function(callback) {
      var currTime = new Date().getTime(),
        timeToCall = Math.max(0, 16 - (currTime - lastTime)),
        id = window.setTimeout(function() {
          callback(currTime + timeToCall);
        }, timeToCall);
      lastTime = currTime + timeToCall;
      return id;
    };

    window.fovCanAnim = function(id) {
      clearTimeout(id);
    };
  }
}

// *** Get a URL parameter by name method.
if (typeof(window.getUrlParam) === 'undefined') {
  window.getUrlParam = function getUrlParam(getParamName) {
    var theUrl = window.location.search.substring(1),
      urlVars = theUrl.split('&');
    for (var i = 0; i < urlVars.length; i++) {
      var paramByName = urlVars[i].split('=');
      if (paramByName[0] == getParamName) {
        return paramByName[1];
      }
    }
  };
}

// *** Convert from degrees to radians.
if (typeof(Number.prototype.toRad) === 'undefined') {
  Number.prototype.toRad = function() {
    return this * Math.PI / 180;
  };
}

// *** Get sign of number, with 0 = +.
if (typeof(Number.prototype.getSign) === 'undefined') {
  Number.prototype.getSign = function() {
    return this > 0 ? 1 : this < 0 ? -1 : 1;
  };
}

// *** Truncate to a whole number without rounding.
if (typeof(Number.prototype.toTrunc) === 'undefined') {
  Number.prototype.toTrunc = function() {
    return Math[this > 0 ? 'floor' : 'ceil'](this);
  };
}

// *** Convert HTML Entity to string Polyfill if String.fromCodePoint not supported by browser.
// *** Public Domain from:
// ***   https://developer.mozilla.org/en/docs/Web/JavaScript/Reference/Global_Objects/String/fromCodePoint
if (!String.fromCodePoint) {
  (function() {
    var defineProperty = (function() {
      var result;
      // IE 8 only supports `Object.defineProperty` on DOM elements
      try {
        var object = {};
        var $defineProperty = Object.defineProperty;
        result = $defineProperty(object, object, object) && $defineProperty;
      } catch (error) {}
      return result;
    }());
    var stringFromCharCode = String.fromCharCode;
    var floor = Math.floor;
    var fromCodePoint = function() {
      var MAX_SIZE = 0x4000;
      var codeUnits = [];
      var highSurrogate;
      var lowSurrogate;
      var index = -1;
      var length = arguments.length;
      if (!length) {
        return '';
      }
      var result = '';
      while (++index < length) {
        var codePoint = Number(arguments[index]);
        if (!isFinite(codePoint) || // `NaN`, `+Infinity`, or `-Infinity`
          codePoint < 0 || // not a valid Unicode code point
          codePoint > 0x10FFFF || // not a valid Unicode code point
          floor(codePoint) != codePoint // not an integer
        ) {
          throw RangeError('Invalid code point: ' + codePoint);
        }
        if (codePoint <= 0xFFFF) { // BMP code point
          codeUnits.push(codePoint);
        } else {
          // Astral code point; split in surrogate halves
          codePoint -= 0x10000;
          highSurrogate = (codePoint >> 10) + 0xD800;
          lowSurrogate = (codePoint % 0x400) + 0xDC00;
          codeUnits.push(highSurrogate, lowSurrogate);
        }
        if (index + 1 == length || codeUnits.length > MAX_SIZE) {
          result += stringFromCharCode.apply(null, codeUnits);
          codeUnits.length = 0;
        }
      }
      return result;
    };
    if (defineProperty) {
      defineProperty(String, 'fromCodePoint', {
        'value': fromCodePoint,
        'configurable': true,
        'writable': true
      });
    } else {
      String.fromCodePoint = fromCodePoint;
    }
  }());
}

// ********** Objects ***************************************************************************************

// *** FOV Reticule constructor.
function fovReticule(wDeg, hDeg, rDeg, wOff, hOff, drawColour, appH, appV, scopeRes, maxOvl, annotateOvl, description) {
  // Reticule details.
  this.wDeg = wDeg;
  this.wOff = wOff;
  this.hDeg = hDeg;
  this.hOff = hOff;
  this.rDeg = rDeg;
  this.drawColour = drawColour;
  this.appH = appH;
  this.appV = appV;
  this.scopeRes = scopeRes;
  this.maxOvl = maxOvl;
  this.annotateOvl = annotateOvl;
  this.description = description;

  // Drawing details.
  this.guid = makeGuid();
  this.aI = null;
  this.animId = 'unset';
  this.forceRedraw = false;
  this.lastSize = [0, 0];
  this.forceZoom = 0;
  this.lastZoom = 0;
  this.foreground = false;

  // Methods.
  this.redrawOn = redrawOn;
  this.redrawOff = redrawOff;
  this.calculateZoom = calculateZoom;
  this.toFore = toFore;
  this.toBack = toBack;
  this.showFov = showFov;
  this.hideFov = hideFov;

  // *** Turn on redrawing of reticule method.
  function redrawOn() {
    // Ensure this function gets called again on each browser animation frame.
    this.animId = fovReqAnim(redrawOn.bind(this));

    // Redraw the reticule if the field of view, Aladin Lite size has changed, or redraw forced (e.g.reticule has rotated, moved or annotation has changed).
    if (this.aI.getFov()[0] != this.lastZoom || this.aI.getSize()[0] != this.lastSize[0] || this.aI.getSize()[1] != this.lastSize[1] || this.forceRedraw) {
      // Reset forced redraw flag.
      this.forceRedraw = false;

      // Set default font style, etc.
      this.aI.ctx[this.guid].font = '10px Hind';
      this.aI.ctx[this.guid].fillStyle = '#E678FA';

      // Clear the canvas.
      this.aI.ctx[this.guid].clearRect(-(this.aI.ctx[this.guid].canvas.width) / 2 - 1, -(this.aI.ctx[this.guid].canvas.height) / 2 - 1,
                                       this.aI.ctx[this.guid].canvas.width + 2, this.aI.ctx[this.guid].canvas.height + 2);

      // Resize the canvas if needed.
      if (this.aI.getSize()[0] != this.lastSize[0] || this.aI.getSize()[1] != this.lastSize[1]) {
        // Resize the canvas to match new view size.
        this.aI.resizeCanvas(this.guid);
        // Recheck whether to maximise reticule or scale the reticule to match the zoom.
        this.calculateZoom();
      }

      // Update Aladin Lite if this is the foreground reticule.
      if (this.foreground) {
        if (this.forceZoom > 0) { // Reticule is maximised.
          // Hide the zoom controls.  Not ideal since it affects all instances of Aladin Lite on the page.
          $('.zoomIn-container, .zoomOut-container').css('display', 'none');
          $('.bws-maxIndicator').css('display', 'block');

          if (this.forceZoom <= 90) { // Reticule bounding box is below 90 degrees.
            // Stop limiting zooming using setFOVRange so that progressiveFov can work.
            this.aI.setFOVRange(null, null);

            // Maximise the reticule but keep it in bounds.
            this.aI.progressiveFov(this.forceZoom, false);

            // Prevent zooming.
            this.aI.setFOVRange(this.aI.getFov()[0], this.aI.getFov()[0]);
          } else { // Reticule bounding box is above 90 degrees.
            // Stop limiting zooming using setFOVRange so that progressiveFov can work.
            this.aI.setFOVRange(null, null);

            // Otherwise zoom to the maximum reliable extent, as close to 90 as possible.
            this.aI.progressiveFov(90, true);

            // Prevent zooming beyond whatever we got close to 90.  Stops FOV 'bouncing' if use tries to zoom.
            this.aI.setFOVRange(this.aI.getFov()[0], this.aI.getFov()[0]);
          }
        } else {
          // Zoomable so show the zoom controls. Not ideal since it affects all instances of Aladin Lite on the page.
          $('.zoomIn-container, .zoomOut-container').css('display', 'block');
          $('.bws-maxIndicator').css('display', 'none');

          // Stop limiting zooming using setFOVRange so that progressiveFov can work.
          this.aI.setFOVRange(null, null);

          if (this.aI.getFov()[0] > 90) {
            // Exceeded maximum reliable Aladin Lite zoom so get as close to 90 as possible.
            this.aI.setFov(1);
            this.aI.progressiveFov(90, true);
          }

          // Limit zooming to between 0 and 90.
          this.aI.setFOVRange(0, 90);
        }
      }

      // Calculate the pixel scale using Aladin Lite's own methods (far more accurate for tiny canvas sizes).
      // Find the change in declination (current line of RA always aligned vertically to frame in Aladin Lite) between centre and bottom of the frame, and divide half the vertical window size by result.
      var scaleDegToPix = Math.round(aladin.getSize()[1]/2) / (aladin.pix2world(Math.round(aladin.getSize()[0]/2)-1, Math.round(aladin.getSize()[1]/2))[1] - aladin.pix2world(Math.round(aladin.getSize()[0]/2)-1, aladin.getSize()[1]-1)[1]);

      // Calculate the size of the FoV reticule by converting from degrees to pixels.
      var rectWidth = this.wDeg * scaleDegToPix,
          rectHeight = this.hDeg * scaleDegToPix;

      // DEBUG: Use Aladin Lite to draw a 1 degree box projected on to the sky for testing.
      // N.B. Vertical axis will be comparable to reticule at any declination.
      // N.B. Horizontal axis only comparable to reticule at the celestial equator as lines of RA converge at the poles.
      // aladin.removeLayers();
      // testOverlay = A.graphicOverlay({color: '#ee2345', lineWidth: 1});
      // aladin.addOverlay(testOverlay);
      // formMain[curRet].gotoRaHD = aladin.view.viewCenter.lon;
      // formMain[curRet].gotoDecDD = aladin.view.viewCenter.lat;
      // testOverlay.add(A.polyline([ [aladin.view.viewCenter.lon-0.5, aladin.view.viewCenter.lat-0.5], [aladin.view.viewCenter.lon+0.5, aladin.view.viewCenter.lat-0.5], [aladin.view.viewCenter.lon+0.5, aladin.view.viewCenter.lat+0.5], [aladin.view.viewCenter.lon-0.5, aladin.view.viewCenter.lat+0.5], [aladin.view.viewCenter.lon-0.5, aladin.view.viewCenter.lat-0.5] ]));

      // Offset the canvas to centre the reticule and rotate it.
      this.aI.ctx[this.guid].translate(this.wOff * scaleDegToPix, this.hOff * scaleDegToPix);
      this.aI.ctx[this.guid].rotate((this.rDeg).toRad());

      // Draw the FoV overlay.
      this.aI.ctx[this.guid].strokeStyle = this.drawColour;
      this.aI.ctx[this.guid].fillStyle = this.drawColour;
      this.aI.ctx[this.guid].strokeRect(-rectWidth / 2, -rectHeight / 2, rectWidth, rectHeight);

      // Web site address if required.
      if (rectWidth > 135) {
        this.aI.ctx[this.guid].font="10px Hind";
        this.aI.ctx[this.guid].fillText('www.blackwaterskies.co.uk', rectWidth / 2 - 130, rectHeight / 2 - 3);
        this.aI.ctx[this.guid].fillText(this.description, rectWidth / 2 - 130, rectHeight / 2 - 13);
      }

      // Annotate the FoV overlay horizontal if required.
      if (this.annotateOvl && rectWidth > 142) {
        if (rectWidth > 249) {
          this.aI.ctx[this.guid].font="12px Hind";
        } else {
          this.aI.ctx[this.guid].font="10px Hind";
        }

        // Width in degrees and horizontal arcseconds per pixel scale.
        this.aI.ctx[this.guid].fillText((this.wDeg).toFixed(2) + '\u00B0 wide ->',
                                        -rectWidth / 2 + 40, -rectHeight / 2 + 13);
        if (this.appH !== 0) {
          this.aI.ctx[this.guid].fillText((this.appH).toFixed(2) + '\u0022 per pixel.',
                                          -rectWidth / 2 + 40, -rectHeight / 2 + 26);
        }

        // Scope resolution (Dawes Limit).
        if (this.scopeRes !== 0) {
          this.aI.ctx[this.guid].fillText('Dawes Limit: ' + (this.scopeRes * 1).toFixed(2) + '\u0022',
                                          -rectWidth / 2 + 40, -rectHeight / 2 + 39);
        }
      }

      // Annotate the FoV overlay vertical if required.
      if (this.annotateOvl && rectHeight > 123) {
        if (rectWidth > 249) {
          this.aI.ctx[this.guid].font = '12px Hind';
        } else {
          this.aI.ctx[this.guid].font = '10px Hind';
        }

        // Height in degrees and vertical arcseconds per pixel scale.
        this.aI.ctx[this.guid].rotate((90).toRad());
        this.aI.ctx[this.guid].fillText((this.hDeg).toFixed(2) + '\u00B0 high ->',
                                        -rectHeight / 2 + 40, rectWidth / 2 - 5);
        if (this.appV !== 0) {
          this.aI.ctx[this.guid].fillText((this.appV).toFixed(2) + '\u0022 per pixel.',
                                          -rectHeight / 2 + 40, rectWidth / 2 - 18);
        }
        this.aI.ctx[this.guid].rotate(-(90).toRad());
      }

      if (this.annotateOvl) {
        // Make Aladin Lite crosshair reticule visible.
        this.aI.view.showReticle(true);
      } else {
        // Hide Aladin Lite crosshair reticule.
        this.aI.view.showReticle(false);
      }

      // Remove canvas rotation and offset.
      this.aI.ctx[this.guid].rotate(-(this.rDeg).toRad());
      this.aI.ctx[this.guid].translate(-this.wOff * scaleDegToPix, -this.hOff * scaleDegToPix);

      // Record the current field of view.
      this.lastZoom = this.aI.getFov()[0];

      // Record the current view size.
      this.lastSize[0] = this.aI.getSize()[0];
      this.lastSize[1] = this.aI.getSize()[1];
    }
  }

  // *** Stop any existing zoom / redraw method.
  function redrawOff() {
    if (this.animId != 'unset') {
      fovCanAnim(this.animId);
      this.animId = 'unset';

      // Remove the overlay layer.
      this.aI.ctx[this.guid].clearRect(-(this.aI.ctx[this.guid].canvas.width) / 2 - 1, -(this.aI.ctx[this.guid].canvas.height) / 2 - 1,
                                       this.aI.ctx[this.guid].canvas.width + 2, this.aI.ctx[this.guid].canvas.height + 2);

      if (this.foreground) {
        // Show the Aladin location Div. Not ideal since it affects all instances of Aladin Lite on the page.
        $('.aladin-location').css('display', 'block');
        $('.zoomIn-container, .zoomOut-container').css('display', 'block');
        $('.bws-maxIndicator').css('display', 'none');

        // Stop limiting zooming using setFOVRange
        this.aI.setFOVRange(null, null);

        // Make Aladin Lite crosshair reticule visible.
        this.aI.view.showReticle(true);
      }
    }
  }

  // *** Check whether to maximise reticule or scale the reticule to match the zoom method.
  function calculateZoom() {
    if (this.maxOvl) {
      // Calculate bounding box height and width of reticule in degrees.
      var bbHeight = boundHeight(-this.wDeg / 2, -this.hDeg / 2, this.wDeg / 2, -this.hDeg / 2,
                                 -this.wDeg / 2, this.hDeg / 2, this.wDeg / 2, this.hDeg / 2, this.rDeg) + Math.abs(this.hOff) * 2,
          bbWidth = boundWidth(-this.wDeg / 2, -this.hDeg / 2, this.wDeg / 2, -this.hDeg / 2,
                               -this.wDeg / 2, this.hDeg / 2, this.wDeg / 2, this.hDeg / 2, this.rDeg) + Math.abs(this.wOff) * 2;

      // Calculate the aspect ratio of the frame.
      var hRatio = this.aI.getSize()[0] / this.aI.getSize()[1];

      // Zoom to the maximum of the bounding box width or bounding box height * the frame ratio, plus 3% to leave a gap around the reticule.
      // N.B. Aladin Lite's zoom is always set in relation to frame width, so we have to scale if the bounding box is taller than wide.
      this.forceZoom = Math.max(bbWidth * 1.03, bbHeight * hRatio * 1.03);
    } else {
      // Let the user zoom Aladin Lite.
      this.forceZoom = 0;
    }
  }

  // *** Move to foreground method.
  function toFore() {
    this.foreground = true;
  }

  // *** Move to background method.
  function toBack() {
    if (this.aI !== null) {
      // Show the Aladin location Div. Not ideal since it affects all instances of Aladin Lite on the page.
      $('.aladin-location').css('display', 'block');
      $('.zoomIn-container, .zoomOut-container').css('display', 'block');
      $('.bws-maxIndicator').css('display', 'none');

      // Stop limiting zooming using setFOVRange
      this.aI.setFOVRange(null, null);

      // Make Aladin Lite crosshair reticule visible.
      this.aI.view.showReticle(true);
    }

    this.foreground = false;
  }

  // *** Show FOV reticule method.
  function showFov(aI) {
    // Stop any previous redrawing.
    this.redrawOff();

    // Record the Aladin Lite instance on which this reticule is being drawn.
    this.aI = aI;

    // Check whether to maximise reticule or scale the reticule to match the zoom.
    this.calculateZoom();

    // Force a redraw of the reticule.
    this.forceRedraw = true;

    // Draw reticule and set animation frame to redraw/zoom as required.
    this.redrawOn();
  }

  // *** Hide FOV reticule method.
  function hideFov() {
    if (this.aI !== null) {
      // Stop any forced zooming or redrawing.
      this.redrawOff();
    }
    this.aI = null;
  }
}

// *** Form UI constructor.
function itForm() {
  this.fovWidth = 2.13;
  this.fovHeight = 1.42;
  this.fovMax = true;
  this.fovNotes = true;
  this.fovColour = '#00FF00';
  this.fovRotation = 0;
  this.fovYOffset = 0;
  this.fovXOffset = 0;
  this.fovAppH = 0;
  this.fovAppV = 0;
  this.fovScopeRes = 0;
  this.compDesc = 'New';
  this.compDisplay = false;
  this.searchTarget = 'Horsehead Nebula';
  this.searchPopular = '';
  this.searchSurveySelection = 'P/DSS2/color-m';
  this.gotoRaH = '';
  this.gotoRaHD = '';
  this.gotoRaM = '';
  this.gotoRaS = '';
  this.gotoAddSign = '';
  this.gotoDecD = '';
  this.gotoDecDD = '';
  this.gotoDecM = '';
  this.gotoDecS = '';
  this.ccdWidth = 22.33;
  this.ccdHeight = 14.89;
  this.ccdFocalLength = 600;
  this.pixelWidth = 4.70;
  this.pixelHeight = 4.70;
  this.pixelHRes = 4752;
  this.pixelVRes = 3168;
  this.pixelFocalLength = 600;
  this.pixelAperture = 80;
  this.equipReducer = 1.00;
  this.equipRawFL = 600;
  this.equipFocalLength = 600;
  this.equipAperture = 80;
  this.equipCameraVal = '{"pixelWidth":"4.7","pixelHeight":"4.7","ccdWidth":"22.33","ccdHeight":"14.89","pixelHRes":"4752","pixelVRes":"3168"}';
  this.equipCameraText = 'Canon EOS-50D / 500D / Rebel T1i';
  this.equipScopeVal = '{"aperture":"80","focalLength":"600"}';
  this.equipScopeText = 'Skywatcher Evostar 80ED DS Pro';

  // Methods.
  this.getAllForm = getAllForm;
  this.putAllForm = putAllForm;
  this.getFovForm = getFovForm;
  this.putFovForm = putFovForm;
  this.populateFovForm = populateFovForm;
  this.buttonFovForm = buttonFovForm;
  this.getCompForm = getCompForm;
  this.putCompForm = putCompForm;
  this.getSearchForm = getSearchForm;
  this.putSearchForm = putSearchForm;
  this.getGotoForm = getGotoForm;
  this.isBlankGotoForm = isBlankGotoForm;
  this.putGotoForm = putGotoForm;
  this.rdsGotoForm = rdsGotoForm;
  this.ddsGotoForm = ddsGotoForm;
  this.clearGotoForm = clearGotoForm;
  this.getCcdForm = getCcdForm;
  this.putCcdForm = putCcdForm;
  this.getPixelForm = getPixelForm;
  this.putPixelForm = putPixelForm;
  this.getEquipForm = getEquipForm;
  this.putEquipForm = putEquipForm;
  this.putUrlForm = putUrlForm;
  this.cleanUrlForm = cleanUrlForm;
  this.dirtyUrlForm = dirtyUrlForm;
  this.createSavedForm = createSavedForm;
  this.retrieveSavedForm = retrieveSavedForm;

  // *** Get all values from form UI method.
  function getAllForm() {
    this.getFovForm();
    this.getCompForm();
    this.getSearchForm();
    this.getGotoForm();
    this.getCcdForm();
    this.getPixelForm();
    this.getEquipForm();
  }

  // *** Update form UI with all values method.
  function putAllForm() {
    this.putFovForm();
    this.putCompForm();
    this.putSearchForm();
    this.putGotoForm();
    this.putCcdForm();
    this.putPixelForm();
    this.putEquipForm();
  }

  // *** Get values from FOV/Comp form UI method.
  function getFovForm() {
    this.fovWidth = Math.abs(textFloat($('#compWidth').val(), 2.13));
    this.fovHeight = Math.abs(textFloat($('#compHeight').val(), 1.42));
    // Bounds checking of width/height depends on the dimensions of the Aladin instance, so do it in UI handlers.
    this.fovMax = $('#compMax').prop('checked');
    this.fovNotes = $('#compNotes').prop('checked');
    this.fovColour = $('#compColour').val();
    this.fovRotation = textInt($('#fovRotation').val(), 0);
    // Bounds check and set the FoV rotation parameter.
    if (formMain.fovRotation < -180 || formMain.fovRotation > 180) {
      formMain.fovRotation = 0;
    }
    this.fovYOffset = textFloat($('#fovYOffset').val(), 0);
    this.fovXOffset = textFloat($('#fovXOffset').val(), 0);
  }

  // *** Update FOV/Comp form UI with values method.
  function putFovForm() {
    $('#compWidth').val((this.fovWidth).toFixed(2));
    $('#compHeight').val((this.fovHeight).toFixed(2));
    $('#fovRotation').val(this.fovRotation);
    $('#fovYOffset').val((this.fovYOffset).toFixed(2));
    $('#fovXOffset').val((this.fovXOffset).toFixed(2));
    $('#compMax').prop('checked', this.fovMax);
    $('#compNotes').prop('checked', this.fovNotes);
    //$('#compColour').jscolor.fromString(this.fovColour); Doesn't work - uses getElementByID though!
    document.getElementById('compColourButton').jscolor.fromString(this.fovColour);
  }

  // *** Populate FOV form from a calculator form.
  function populateFovForm(chipW, chipH, focalLength, hRes, vRes, aperture) {
    // Convert the chip size in mm to FoV in degrees (small angle approximation).
    // this.fovWidth = chipW / focalLength * 57.3;
    // this.fovHeight = chipH / focalLength * 57.3;
    // Convert the chip size in mm to FoV in degrees (arctan version - more accurate for large chips/short focal lengths).
    this.fovWidth = Math.atan(chipW / focalLength) * 57.2957795;
    this.fovHeight = Math.atan(chipH / focalLength) * 57.2957795;

    // If the chip resolution is known, calculate the pixel scale in arcseconds per pixel.
    if (hRes > 0 && vRes > 0) {
      this.fovAppH = (this.fovWidth * 3600) / hRes;
      this.fovAppV = (this.fovHeight * 3600) / vRes;
    } else {
      this.fovAppH = 0;
      this.fovAppV = 0;
    }

    // If the scope aperture is known, calculate its resolving power (Dawes Limit).
    if (aperture > 0) {
      this.fovScopeRes = 116 / aperture;
    } else {
      this.fovScopeRes = 0;
    }
  }

  // *** Update FOV show button method.
  function buttonFovForm(theLabel) {
    $('#fovShow').prop('value', theLabel);
  }

  // *** Get values from Compare form UI method.
  function getCompForm() {
    this.compDesc = $('#compDesc').val();
  }

  // *** Update Compare form UI with values method.
  function putCompForm() {
    $('#compDesc').val(this.compDesc);
  }

  // *** Get values from Search form UI method.
  function getSearchForm() {
    this.searchTarget = $('#searchTarget').val();
    this.searchPopular = $('#searchPopular').val();
    this.searchSurveySelection = $('#searchSurveySelection').val();
  }

  // *** Update Search form UI with values method.
  function putSearchForm() {
    $('#searchTarget').val(this.searchTarget);
    $('#searchPopular').val(this.searchPopular);
    $('#searchSurveySelection').val(this.searchSurveySelection);
  }

  // *** Get values from Goto form UI method.
  function getGotoForm() {
    this.gotoRaH = Math.abs(textInt($('#gotoRaH').val(), 0));
    this.gotoRaHD = Math.abs(textFloat($('#gotoRaH').val(), 0));
    this.gotoRaM = Math.abs(textInt($('#gotoRaM').val(), 0));
    this.gotoRaS = Math.abs(textFloat($('#gotoRaS').val(), 0));
    this.gotoDecD = textInt($('#gotoDecD').val(), 0);
    this.gotoDecDD = textFloat($('#gotoDecD').val(), 0);
    this.gotoDecM = Math.abs(textInt($('#gotoDecM').val(), 0));
    this.gotoDecS = Math.abs(textFloat($('#gotoDecS').val(), 0));
    // Check for negative Dec sign for negative sexagesimal values smaller than one degree.
    if (($.trim($('#gotoDecD').val())).substring(0, 1) == '-') {
      this.gotoAddSign = '-';
    } else {
      this.gotoAddSign = '';
    }

    // Check and process decimal or sexagesimal RA.
    if (!$.isNumeric($('#gotoRaM').val()) && !$.isNumeric($('#gotoRaS').val())) {
      // Use the decimal RA in the first input box.
      if (this.gotoRaHD < 0) {
        // Decimal RA out of bounds.
        this.gotoRaH = 0;
        this.gotoRaM = '';
        this.gotoRaS = '';
      } else if (this.gotoRaHD >= 360) {
        // Decimal RA out of bounds.
        this.gotoRaH = 359.99;
        this.gotoRaM = '';
        this.gotoRaS = '';
      }
      // Convert decimal RA to sexagesimal.
      this.rdsGotoForm();
    } else {
      // Bounds check sexagesimal RA.
      if (this.gotoRaH < 0) {
        this.gotoRaH = 0;
        this.gotoRaM = 0;
        this.gotoRaS = 0;
      } else if (this.gotoRaH > 23) {
        this.gotoRaH = 23;
        this.gotoRaM = 59;
        this.gotoRaS = 59.99;
      } else if (this.gotoRaM < 0) {
        this.gotoRaM = 0;
        this.gotoRaS = 0;
      } else if (this.gotoRaM > 59) {
        this.gotoRaM = 59;
        this.gotoRaS = 59.99;
      } else if (this.gotoRaS < 0) {
        this.gotoRaS = 0;
      } else if (this.gotoRaS >= 60) {
        this.gotoRaS = 59.99;
      }
    }

    // Check and process decimal or sexagesimal Dec.
    if ((!$.isNumeric($('#gotoDecM').val()) && !$.isNumeric($('#gotoDecS').val()))) {
      // Use the decimal Dec in the first input box.
      if (this.gotoDecDD < -90) {
        // Decimal Dec out of bounds.
        this.gotoDecD = -90;
        this.gotoDecM = '';
        this.gotoDecS = '';
        this.gotoAddSign = '-';
      } else if (this.gotoDecDD > 90) {
        // Decimal Dec out of bounds.
        this.gotoDecD = 90;
        this.gotoDecM = '';
        this.gotoDecS = '';
        this.gotoAddSign = '';
      }
      // Convert decimal Dec to sexagesimal.
      this.ddsGotoForm();
    } else {
      // Bounds check sexagesimal Dec.
      if (this.gotoDecD < -90) {
        this.gotoDecD = -90;
        this.gotoDecM = 0;
        this.gotoDecS = 0;
        this.gotoAddSign = '-';
      } else if (this.gotoDecD > 90) {
        this.gotoDecD = 90;
        this.gotoDecM = 0;
        this.gotoDecS = 0;
        this.gotoAddSign = '';
      } else if (this.gotoDecD == -90 || this.gotoDecD == 90) {
        this.gotoDecM = 0;
        this.gotoDecS = 0;
      } else {
        if (this.gotoDecM < 0) {
          this.gotoDecM = 0;
          this.gotoDecS = 0;
        } else if (this.gotoDecM > 59) {
          this.gotoDecM = 59;
          this.gotoDecS = 59.99;
        } else if (this.gotoDecS < 0) {
          this.gotoDecS = 0;
        } else if (this.gotoDecS >= 60) {
          this.gotoDecS = 59.99;
        }
      }
    }

    // Recalculate decimal RA and Dec from validated / converted sexagesimal RA and Dec.
    this.gotoRaHD = this.gotoRaH * 15 + this.gotoRaM / 4 + this.gotoRaS / 240;
    this.gotoDecDD = (Math.abs(this.gotoDecD) + this.gotoDecM / 60 + this.gotoDecS / 3600) * ((this.gotoAddSign == '-') ? -1 : 1);
  }

  // *** Check values from Goto form UI method.
  function isBlankGotoForm() {
    return $('#gotoRaH').val() === '' && $('#gotoDecD').val() === '';
  }

  // *** Update Goto form UI with values method.
  function putGotoForm() {
    if (this.gotoRaH !== '') {
      $('#gotoRaH').val((this.gotoRaH).toFixed());
      $('#gotoRaM').val((this.gotoRaM).toFixed());
      $('#gotoRaS').val((this.gotoRaS).toFixed(2));
      $('#gotoDecD').val(((this.gotoDecD === 0) ? this.gotoAddSign : '') + (this.gotoDecD * 1).toFixed());
      $('#gotoDecM').val((this.gotoDecM).toFixed());
      $('#gotoDecS').val((this.gotoDecS).toFixed(2));
    } else {
      $('#gotoRaH').val('');
      $('#gotoRaM').val('');
      $('#gotoRaS').val('');
      $('#gotoDecD').val('');
      $('#gotoDecM').val('');
      $('#gotoDecS').val('');
    }
  }

  // *** Convert decimal RA to sexagesimal method.
  function rdsGotoForm() {
    // Convert decimal RA to sexagesimal.
    this.gotoRaH = (this.gotoRaHD / 15).toTrunc();
    this.gotoRaM = ((this.gotoRaHD - this.gotoRaH * 15) * 4).toTrunc();
    this.gotoRaS = (this.gotoRaHD - this.gotoRaH * 15 - this.gotoRaM / 4) * 240;
  }

  // *** Convert decimal Dec to sexagesimal method.
  function ddsGotoForm() {
    // Convert decimal Dec to sexagesimal.
    this.gotoDecD = (this.gotoDecDD).toTrunc();
    this.gotoDecM = ((this.gotoDecDD - this.gotoDecD) * 60).toTrunc();
    this.gotoDecS = Math.abs(((this.gotoDecDD - this.gotoDecD - this.gotoDecM / 60) * 3600));
    this.gotoDecM = Math.abs(this.gotoDecM);

    // Deal with negative Dec values that are smaller than 1 degree.
    if (this.gotoDecD === 0 && this.gotoDecDD < 0) {
      this.gotoAddSign = '-';
    } else {
      this.gotoAddSign = '';
    }
  }

  // *** Clear Goto form values.
  function clearGotoForm() {
    this.gotoRaH = '';
    this.gotoRaHD = '';
    this.gotoRaM = '';
    this.gotoRaS = '';
    this.gotoAddSign = '';
    this.gotoDecD = '';
    this.gotoDecDD = '';
    this.gotoDecM = '';
    this.gotoDecS = '';
  }

  // *** Get values from CCD form UI method.
  function getCcdForm() {
    this.ccdWidth = (Math.abs(textFloat($('#ccdWidth').val(), 22.33)));
    this.ccdHeight = (Math.abs(textFloat($('#ccdHeight').val(), 14.89)));
    this.ccdFocalLength = (Math.abs(textInt($('#ccdFocalLength').val(), 600)));
  }

  // *** Update CCD form UI with values method.
  function putCcdForm() {
    $('#ccdWidth').val((this.ccdWidth).toFixed(2));
    $('#ccdHeight').val((this.ccdHeight).toFixed(2));
    $('#ccdFocalLength').val((this.ccdFocalLength).toFixed());
  }

  // *** Get values from Pixel form UI method.
  function getPixelForm() {
    this.pixelWidth = (Math.abs(textFloat($('#pixelWidth').val(), 4.7)));
    this.pixelHeight = (Math.abs(textFloat($('#pixelHeight').val(), 4.7)));
    this.pixelHRes = (Math.abs(textInt($('#pixelHRes').val(), 4770)));
    this.pixelVRes = (Math.abs(textInt($('#pixelVRes').val(), 3178)));
    this.pixelFocalLength = (Math.abs(textInt($('#pixelFocalLength').val(), 600)));
    this.pixelAperture = (Math.abs(textInt($('#pixelAperture').val(), 80)));
  }

  // *** Update Pixel form UI with values method.
  function putPixelForm() {
    $('#pixelWidth').val((this.pixelWidth).toFixed(2));
    $('#pixelHeight').val((this.pixelHeight).toFixed(2));
    $('#pixelHRes').val((this.pixelHRes).toFixed());
    $('#pixelVRes').val((this.pixelVRes).toFixed());
    $('#pixelFocalLength').val((this.pixelFocalLength).toFixed());
    $('#pixelAperture').val((this.pixelAperture).toFixed());
  }

  // *** Get values from Equipment form UI method.
  function getEquipForm() {
    this.equipCameraVal = $('#equipCamera').val();
    this.equipCameraText = $('#equipCamera option').filter(':selected').text();
    this.equipScopeVal = $('#equipScope').val();
    this.equipScopeText = $('#equipScope option').filter(':selected').text();
    this.equipReducer = (Math.abs(textFloat($('#equipReducer').val(), 1)));

    // Parse the camera and scope parameters from the selects. N.B. Need to convert strings to numbers with * 1.
    var tempCamera = window.JSON.parse(this.equipCameraVal),
      tempScope = window.JSON.parse(this.equipScopeVal);
    this.equipFocalLength = tempScope.focalLength * this.equipReducer;
    this.equipAperture = tempScope.aperture * 1;
    this.equipRawFL = tempScope.focalLength * 1;

    // Populate other forms with the relevant values from the equipment.
    this.ccdWidth = tempCamera.ccdWidth * 1;
    this.ccdHeight = tempCamera.ccdHeight * 1;
    this.pixelWidth = tempCamera.pixelWidth * 1;
    this.pixelHeight = tempCamera.pixelHeight * 1;
    this.pixelHRes = tempCamera.pixelHRes * 1;
    this.pixelVRes = tempCamera.pixelVRes * 1;

    // Update other forms with required values. N.B. Exceptionally done in the forms object, normally done in UI handlers.
    this.pixelFocalLength = this.equipFocalLength;
    this.ccdFocalLength = this.equipFocalLength;
    this.pixelAperture = this.equipAperture;
  }

  // *** Update Equipment form UI with values method.
  function putEquipForm() {
    $('#equipRawFL').val((this.equipRawFL).toFixed());
    $('#equipReducer').val((this.equipReducer).toFixed(2));
    $('#equipFocalLength').val((this.equipFocalLength).toFixed());
    $('#equipAperture').val((this.equipAperture).toFixed());

    // De-select the current camera.
    $('#equipCamera option').filter(':selected').prop('selected', false);

    // Check whether retrieved camera is in the list (deals with saved forms that might have old settings).
    var tempCameraText = this.equipCameraText,
      cameraInList = $('#equipCamera option').filter(function() {
        return ($(this).text() == tempCameraText);
      });
    if (cameraInList.length) {
      // Camera still exists so select it.
      cameraInList.prop('selected', true);
    } else {
      // Camera no longer exists so add it.
      $('#equipCamera').append($('<option></option>')
        .prop('value', this.equipCameraVal)
        .text(this.equipCameraText)
        .prop('selected', true)
      );
    }

    // De-select the current scope.
    $('#equipScope option').filter(':selected').prop('selected', false);

    // Check whether retrieved scope is in the list (deals with saved forms that might have old settings).
    var tempScopeText = this.equipScopeText,
      scopeInList = $('#equipScope option').filter(function() {
        return ($(this).text() == tempScopeText);
      });
    if (scopeInList.length) {
      // Scope still exists so select it.
      scopeInList.prop('selected', true);
    } else {
      // Scope no longer exists so add it.
      $('#equipScope').append($('<option></option>')
        .prop('value', this.equipScopeVal)
        .text(this.equipScopeText)
        .prop('selected', true)
      );
    }
  }

  // *** Update Url form UI with values method.
  function putUrlForm(theUrl) {
    $('#urlSave').val(theUrl);
  }

  // *** Update Url form UI to show all reticules saved.
  function cleanUrlForm() {
    $('#urlStore').prop('value', 'All Reticules Saved');
  }

  // *** Update Url form UI to show unsaved reticules.
  function dirtyUrlForm() {
    $('#urlStore').prop('value', 'Save Reticules');
  }

  // *** Create URL from the forms and displayed FOV.
  function createSavedForm(sourceFov) {
    // Create object with all required form values (no methods, as we have to serialise this so keep it small).
    var saveForm = {};
    saveForm.version = 3;

    // Version 3 properties.
    saveForm.fovMax = this.fovMax;
    saveForm.fovNotes = this.fovNotes;

    // Version 2 properties.
    // saveForm.fovStyle = this.fovStyle; // Obsolete as of Version 3
    saveForm.fovYOffset = this.fovYOffset;
    saveForm.fovXOffset = this.fovXOffset;
    saveForm.compDesc = this.compDesc;
    saveForm.compDisplay = this.compDisplay;
    saveForm.searchTarget = this.searchTarget;
    saveForm.searchPopular = this.searchPopular;
    saveForm.gotoRaHD = this.gotoRaHD;
    saveForm.gotoAddSign = this.gotoAddSign;
    saveForm.gotoDecDD = this.gotoDecDD;

    // Version 1 properties.
    saveForm.fovWidth = this.fovWidth;
    saveForm.fovHeight = this.fovHeight;
    saveForm.fovColour = this.fovColour;
    saveForm.fovRotation = this.fovRotation;
    saveForm.fovAppH = this.fovAppH;
    saveForm.fovAppV = this.fovAppV;
    saveForm.fovScopeRes = this.fovScopeRes;
    saveForm.searchSurveySelection = this.searchSurveySelection;
    saveForm.gotoRaH = this.gotoRaH;
    saveForm.gotoRaM = this.gotoRaM;
    saveForm.gotoRaS = this.gotoRaS;
    saveForm.gotoDecD = this.gotoDecD;
    saveForm.gotoDecM = this.gotoDecM;
    saveForm.gotoDecS = this.gotoDecS;
    saveForm.ccdWidth = this.ccdWidth;
    saveForm.ccdHeight = this.ccdHeight;
    saveForm.ccdFocalLength = this.ccdFocalLength;
    saveForm.pixelWidth = this.pixelWidth;
    saveForm.pixelHeight = this.pixelHeight;
    saveForm.pixelHRes = this.pixelHRes;
    saveForm.pixelVRes = this.pixelVRes;
    saveForm.pixelFocalLength = this.pixelFocalLength;
    saveForm.pixelAperture = this.pixelAperture;
    saveForm.equipReducer = this.equipReducer;
    saveForm.equipRawFL = this.equipRawFL;
    saveForm.equipFocalLength = this.equipFocalLength;
    saveForm.equipAperture = this.equipAperture;
    saveForm.equipCameraVal = this.equipCameraVal;
    saveForm.equipCameraText = this.equipCameraText;
    saveForm.equipScopeVal = this.equipScopeVal;
    saveForm.equipScopeText = this.equipScopeText;
    if (sourceFov.aI === null) {
      saveForm.zoom = 2.19;
    } else {
      saveForm.zoom = sourceFov.aI.getFov()[0];
    }

    return saveForm;
  }

  // *** Populate form from saved form object (there are old version 1 saves and new version 2 saves out there).
  function retrieveSavedForm(retForm) {
    // Populate form with saved values.
    if (retForm.version == 1 || retForm.version == 2 || retForm.version == 3) {
      // Version 1 properties.  *** Coerce version 1 saved text values to numbers by * 1.
      this.fovWidth = retForm.fovWidth * 1;
      this.fovHeight = retForm.fovHeight * 1;
      this.fovColour = retForm.fovColour;
      this.fovRotation = retForm.fovRotation * 1;
      this.fovAppH = retForm.fovAppH * 1;
      this.fovAppV = retForm.fovAppV * 1;
      this.fovScopeRes = retForm.fovScopeRes * 1;
      // Make sure the selected survey is still available, otherwise default to DSS coloured.
      if (this.searchSurveySelection == 'P/DSS2/color' ||
        this.searchSurveySelection == 'P/Fermi/color' ||
        this.searchSurveySelection == 'P/XMM/EPIC' ||
        this.searchSurveySelection == 'P/GALEXGR6/AIS/color' ||
        this.searchSurveySelection == 'P/DSS2/red' ||
        this.searchSurveySelection == 'P/SDSS9/color' ||
        this.searchSurveySelection == 'P/Mellinger/color' ||
        this.searchSurveySelection == 'P/IRIS/color' ||
        this.searchSurveySelection == 'P/SPITZER/color' ||
        this.searchSurveySelection == 'P/2MASS/color' ||
        this.searchSurveySelection == 'P/Finkbeiner' ||
        this.searchSurveySelection == 'P/VTSS/Ha'
      ) {
        this.searchSurveySelection = retForm.searchSurveySelection;
      } else {
        this.searchSurveySelection = 'P/DSS2/color';
      }
      this.gotoRaH = retForm.gotoRaH * 1;
      this.gotoRaM = retForm.gotoRaM * 1;
      this.gotoRaS = retForm.gotoRaS * 1;
      this.gotoDecD = retForm.gotoDecD * 1;
      this.gotoDecM = retForm.gotoDecM * 1;
      this.gotoDecS = retForm.gotoDecS * 1;
      this.ccdWidth = retForm.ccdWidth * 1;
      this.ccdHeight = retForm.ccdHeight * 1;
      this.ccdFocalLength = retForm.ccdFocalLength * 1;
      this.pixelWidth = retForm.pixelWidth * 1;
      this.pixelHeight = retForm.pixelHeight * 1;
      this.pixelHRes = retForm.pixelHRes * 1;
      this.pixelVRes = retForm.pixelVRes * 1;
      this.pixelFocalLength = retForm.pixelFocalLength * 1;
      this.pixelAperture = retForm.pixelAperture * 1;
      this.equipReducer = retForm.equipReducer * 1;
      this.equipRawFL = retForm.equipRawFL * 1;
      this.equipFocalLength = retForm.equipFocalLength * 1;
      this.equipAperture = retForm.equipAperture * 1;
      this.equipCameraVal = retForm.equipCameraVal;
      this.equipCameraText = retForm.equipCameraText;
      this.equipScopeVal = retForm.equipScopeVal;
      this.equipScopeText = retForm.equipScopeText;

      if (retForm.version == 3) {
        // Version 3 properties.
        this.fovYOffset = retForm.fovYOffset;
        this.fovXOffset = retForm.fovXOffset;
        this.compDesc = retForm.compDesc;
        this.compDisplay = retForm.compDisplay;
        this.searchTarget = retForm.searchTarget;
        this.searchPopular = retForm.searchPopular;
        this.gotoRaHD = retForm.gotoRaHD;
        this.gotoAddSign = retForm.gotoAddSign;
        this.gotoDecDD = retForm.gotoDecDD;
        this.fovMax = retForm.fovMax;
        this.fovNotes = retForm.fovNotes;
      } else if (retForm.version == 2) {
        // Version 2 properties.
        this.fovYOffset = retForm.fovYOffset;
        this.fovXOffset = retForm.fovXOffset;
        this.compDesc = retForm.compDesc;
        this.compDisplay = retForm.compDisplay;
        this.searchTarget = retForm.searchTarget;
        this.searchPopular = retForm.searchPopular;
        this.gotoRaHD = retForm.gotoRaHD;
        this.gotoAddSign = retForm.gotoAddSign;
        this.gotoDecDD = retForm.gotoDecDD;
        // Derive version 3 properties from version 2 form.
        this.fovMax = (retForm.fovStyle == 'M' || retForm.fovStyle == 'MA');
        this.fovNotes = (retForm.fovStyle == 'ZA' || retForm.fovStyle == 'MA');
      } else {
        // Derive version 3 properties from version 1 form.
        this.fovYOffset = 0;
        this.fovXOffset = 0;
        this.searchTarget = '';
        this.searchPopular = '';
        this.compDesc = 'Description';
        this.compDisplay = retForm.showReticule;

        // Check for negative Dec sign for negative sexagesimal values smaller than one degree.
        if ($.trim(retForm.gotDecD).substring(0, 1) == '-') {
          this.gotoAddSign = '-';
        } else {
          this.gotoAddSign = '';
        }

        // Calculate decimal RA and Dec from saved sexagesimal RA and Dec.
        this.gotoRaHD = this.gotoRaH * 15 + this.gotoRaM / 4 + this.gotoRaS / 240;
        this.gotoDecDD = (Math.abs(this.gotoDecD) + this.gotoDecM / 60 + this.gotoDecS / 3600) * ((this.gotoAddSign == '-') ? -1 : 1);

        // Derive version 3 properties from version 1 form.
        this.fovMax = retForm.fovMax;
        this.fovNotes = true;
      }
    } else {
      // Unrecognised URL version.
    }
  }
}

// ********** Utility Functions *******************************************************************************

// *** Convert text input to integer.
function textInt(textVal, defaultVal) {
  if (isNaN(parseInt(textVal, 10))) {
    return defaultVal;
  } else {
    return parseInt(textVal, 10);
  }
}

// *** Convert text input to float.
function textFloat(textVal, defaultVal) {
  if (isNaN(parseFloat(textVal))) {
    return defaultVal;
  } else {
    return parseFloat(textVal);
  }
}

// *** Calculate the width of bounding box after rotating the box specified by XY pairs around origin by Theta.  All parameters/return in degrees.
function boundWidth(xTL, yTL, xTR, yTR, xBL, yBL, xBR, yBR, theta) {
  var cT = Math.cos((theta).toRad()),
    sT = Math.sin((theta).toRad()),
    xTLr = xTL * cT + yTL * sT,
    xTRr = xTR * cT + yTR * sT,
    xBLr = xBL * cT + yBL * sT,
    xBRr = xBR * cT + yBR * sT;
  return Math.max(xTLr, xTRr, xBLr, xBRr) - Math.min(xTLr, xTRr, xBLr, xBRr);
}

// *** Calculate the height of bounding box after rotating the box specified by XY pairs around origin by Theta.  All parameters/return in degrees
function boundHeight(xTL, yTL, xTR, yTR, xBL, yBL, xBR, yBR, theta) {
  var cT = Math.cos((theta).toRad()),
    sT = Math.sin((theta).toRad()),
    yTLr = xTL * sT + yTL * cT,
    yTRr = xTR * sT + yTR * cT,
    yBLr = xBL * sT + yBL * cT,
    yBRr = xBR * sT + yBR * cT;
  return Math.max(yTLr, yTRr, yBLr, yBRr) - Math.min(yTLr, yTRr, yBLr, yBRr);
}

// *** Return a GUID.
function makeGuid(a) {
  return a ? (a ^ Math.random() * 16 >> a / 4).toString(16) : ([1e7] + -1e3 + -4e3 + -8e3 + -1e11).replace(/[018]/g, makeGuid);
}

// ********** Form UI *****************************************************************************************

// *** Toggle visibility of main UI form with formID, hide all others.
function toggleForm(formID) {
  // Determine the final state of the targeted form.
  var toggleToState = $(formID).css('display') == 'none' ? 'block' : 'none';

  // Hide all forms as only one should be visible at any time.
  $('#equipForm, #pixelForm, #ccdForm, #compForm, #searchForm, #gotoForm, #urlForm').css('display', 'none');

  // Set the visibility of the targeted form to the desired state.
  $(formID).css('display', toggleToState);
}

function initUI() { // ****** Activation Section Start.

  // ******* Main UI Buttons

  $('#equip, #equipFormClose').on('keyup click', function() {
    // Toggle visibility of form that corresponds to clicked button.
    toggleForm('#equipForm');
  });

  $('#pixel, #pixelFormClose').on('keyup click', function() {
    // Toggle visibility of form that corresponds to clicked button.
    toggleForm('#pixelForm');
  });

  $('#ccd, #ccdFormClose').on('keyup click', function() {
    // Toggle visibility of form that corresponds to clicked button.
    toggleForm('#ccdForm');
  });

  $('#comp, #compFormClose').on('keyup click', function() {
    // Toggle visibility of form that corresponds to clicked button.
    toggleForm('#compForm');
  });

  $('#search, #searchFormClose').on('keyup click', function() {
    // Toggle visibility of form that corresponds to clicked button.
    toggleForm('#searchForm');
  });

  $('#goto, #gotoFormClose').on('keyup click', function() {
    // Toggle visibility of form that corresponds to clicked button.
    toggleForm('#gotoForm');
  });

  $('#url, #urlFormClose').on('keyup click', function() {
    // Toggle visibility of form that corresponds to clicked button.
    toggleForm('#urlForm');
  });

  $('#maximise').on('keyup click', function() {
    // Toggle between minimised and maximised.
    aladin.toggleFullscreen();

    // Update the UI.
    if ($('#maximise').text() == String.fromCodePoint(0xE5D0)) {
      $('#maximise').text(String.fromCodePoint(0xE5D1));
      $('#maximise').prop('title', 'Exit full screen');

      // Go full-screen
      var fullScr = document.getElementById("aladin-lite-div");
      if (fullScr.requestFullscreen) {
        fullScr.requestFullscreen();
      } else if (fullScr.webkitRequestFullscreen) {
        fullScr.webkitRequestFullscreen();
      } else if (fullScr.mozRequestFullScreen) {
        fullScr.mozRequestFullScreen();
      } else if (fullScr.msRequestFullscreen) {
        fullScr.msRequestFullscreen();
      }
    } else {
      // Exit full-screen
      if (document.exitFullscreen) {
        document.exitFullscreen();
      } else if (document.webkitExitFullscreen) {
        document.webkitExitFullscreen();
      } else if (document.mozCancelFullScreen) {
        document.mozCancelFullScreen();
      } else if (document.msExitFullscreen) {
        document.msExitFullscreen();
      }

      $('#maximise').text(String.fromCodePoint(0xE5D0));
      $('#maximise').prop('title', 'Full screen');
    }
  });

  $('#constellation').on('keyup click', function() {
    // Display or hide constellation outlines and names.
    constellationsVisible = !constellationsVisible;
    if (constellationsVisible) {
      constellationLinesOverlay.show();
      constellationNamesCatalog.show();
    } else {
      constellationLinesOverlay.hide();
      constellationNamesCatalog.hide();
    }

    // Update the UI.
    if ($('#constellation').text() == String.fromCodePoint(0xE83A)) {
      $('#constellation').text(String.fromCodePoint(0xE838));
      $('#constellation').prop('title', 'Show constellations');
    } else {
      $('#constellation').text(String.fromCodePoint(0xE83A));
      $('#constellation').prop('title', 'Hide constellations');
    }
  });

  $('#grid').on('keyup click', function() {
    // Display or hide equatorial grid.
    equatorialGridVisible = !equatorialGridVisible;
    if (equatorialGridVisible) {
      equatorOverlay.show();
      equatorCatalog.show();
      raLinesOverlay.show();
      raLinesCatalog.show();
      decLinesOverlay.show();
      decLinesCatalog.show();
      cardinalOverlay.show();
      cardinalCatalog.show();
    } else {
      equatorOverlay.hide();
      equatorCatalog.hide();
      raLinesOverlay.hide();
      raLinesCatalog.hide();
      decLinesOverlay.hide();
      decLinesCatalog.hide();
      cardinalOverlay.hide();
      cardinalCatalog.hide();
    }

    // Update the UI.
    if ($('#grid').text() == String.fromCodePoint(0xE3EB)) {
      $('#grid').text(String.fromCodePoint(0xE3EC));
      $('#grid').prop('title', 'Show equatorial grid');
    } else {
      $('#grid').text(String.fromCodePoint(0xE3EB));
      $('#grid').prop('title', 'Hide equatorial grid');
    }
  });

  $('#targets').on('keyup click', function() {
    // Display or hide popular target catalogs.
    messierVisible = !messierVisible;
    if (messierVisible) {
      messierCatalog.show();
    } else {
      messierCatalog.hide();
    }

    caldwellVisible = !caldwellVisible;
    if (caldwellVisible) {
      caldwellCatalog.show();
    } else {
      caldwellCatalog.hide();
    }

    // Update the UI.
    if ($('#targets').text() == String.fromCodePoint(0xE148)) {
      $('#targets').text(String.fromCodePoint(0xE147));
      $('#targets').prop('title', 'Show popular targets (Messier, etc.)');
    } else {
      $('#targets').text(String.fromCodePoint(0xE148));
      $('#targets').prop('title', 'Hide popular targets (Messier, etc.)');
    }
  });

  $('#help').on('keyup click', function() {
    // Open help page.
    window.open('/imaging-toolbox-instructions/','_blank');
  });

  $('#zoomIn').on('keyup click', function() {
    // Zoom in.
    aladin.increaseZoom(1);
  });

  $('#zoomOut').on('keyup click', function() {
    // Zoom out.
    aladin.decreaseZoom(1);
  });

  $('#equip, #pixel, #ccd, #comp, #search, #goto, #url, #zoomIn, #zoomOut, #maximise').bind('mousedown', function(prevent) {
    // Prevent text selection.
    prevent.preventDefault();
  });

  // ******* FOV Form

  // *** Respond to Compare Prev button.
  $('#fovPrev').on('keyup click', function() {
    // Display previous reticule if not at first one.
    if (curRet > 0) {
      // Make the current reticule background.
      fovMain[curRet].toBack();

      curRet = curRet - 1;

      // Update the form UI with the values used.
      formMain[curRet].putAllForm();
      compNavigate();

      if (formMain[curRet].compDisplay) {
        // Show the FoV reticule.
        fovButtonShow();
      } else {
        // Hide the FoV reticule (only effect is to update FOV form UI).
        fovButtonHide();
      }
    }
  });

  // *** Respond to Compare Next button.
  $('#fovNext').on('keyup click', function() {
    // Display next reticule if not at last one.
    if (curRet < formMain.length - 1) {
      // Make the current reticule background.
      fovMain[curRet].toBack();

      curRet = curRet + 1;

      // Update the form UI with the values used.
      formMain[curRet].putAllForm();
      compNavigate();

      if (formMain[curRet].compDisplay) {
        // Show the FoV reticule.
        fovButtonShow();
      } else {
        // Hide the FoV reticule (only effect is to update FOV form UI).
        fovButtonHide();
      }
    }
  });

  // *** Respond to FOV Rotate - 10 button.
  $('#fovRotateMinusTen').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Bounds check and decrement the FoV rotation parameter.
    if (formMain[curRet].fovRotation > -171) {
      formMain[curRet].fovRotation = formMain[curRet].fovRotation - 10;
    } else {
      formMain[curRet].fovRotation = -180;
    }

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to FOV Rotate - 1 button.
  $('#fovRotateMinusOne').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Bounds check and decrement the FoV rotation parameter.
    if (formMain[curRet].fovRotation > -179) {
      formMain[curRet].fovRotation = formMain[curRet].fovRotation - 1;
    } else {
      formMain[curRet].fovRotation = -180;
    }

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to change of FOV Rotation value.
  $('#fovRotation').on('change', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to FOV Rotate +1 button.
  $('#fovRotatePlusOne').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Bounds check and increment the FoV rotation parameter.
    if (formMain[curRet].fovRotation < 180) {
      formMain[curRet].fovRotation = formMain[curRet].fovRotation + 1;
    } else {
      formMain[curRet].fovRotation = 180;
    }

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to FOV Rotate +10 button.
  $('#fovRotatePlusTen').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Bounds check and increment the FoV rotation parameter.
    if (formMain[curRet].fovRotation < 171) {
      formMain[curRet].fovRotation = formMain[curRet].fovRotation + 10;
    } else {
      formMain[curRet].fovRotation = 180;
    }

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to FOV Y Up button.
  $('#fovYUp').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Decrement the Y offset parameter.
    formMain[curRet].fovYOffset = formMain[curRet].fovYOffset - 0.05;

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to change of FOV Y Offset value.
  $('#fovYOffset').on('change', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to FOV Y Down button.
  $('#fovYDown').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Increment the Y offset parameter.
    formMain[curRet].fovYOffset = formMain[curRet].fovYOffset + 0.05;

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to FOV X Left button.
  $('#fovXLeft').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Decrement the X offset parameter.
    formMain[curRet].fovXOffset = formMain[curRet].fovXOffset - 0.05;

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to change of FOV X Offset value.
  $('#fovXOffset').on('change', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to FOV X Right button.
  $('#fovXRight').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();

    // Increment the X offset parameter.
    formMain[curRet].fovXOffset = formMain[curRet].fovXOffset + 0.05;

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // Show current reticule and update button UI.
  function fovButtonShow() {
    // Show current reticule.
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getFovForm();
    formMain[curRet].compDisplay = true;

    // Populate the FOV reticule with retrieved values.
    fovMain[curRet].wDeg = formMain[curRet].fovWidth;
    fovMain[curRet].hDeg = formMain[curRet].fovHeight;
    fovMain[curRet].rDeg = formMain[curRet].fovRotation;
    fovMain[curRet].hOff = formMain[curRet].fovYOffset;
    fovMain[curRet].wOff = formMain[curRet].fovXOffset;
    fovMain[curRet].maxOvl = formMain[curRet].fovMax;
    fovMain[curRet].annotateOvl = formMain[curRet].fovNotes;
    fovMain[curRet].drawColour = formMain[curRet].fovColour;
    fovMain[curRet].appH = formMain[curRet].fovAppH;
    fovMain[curRet].appV = formMain[curRet].fovAppV;
    fovMain[curRet].scopeRes = formMain[curRet].fovScopeRes;
    fovMain[curRet].description = formMain[curRet].compDesc;

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();

    // Make the current reticule foreground.
    fovMain[curRet].toFore();

    // Display the FOV overlay.
    fovMain[curRet].showFov(aladin);

    // Update the form UI.
    formMain[curRet].buttonFovForm(String.fromCodePoint(0xE8F5));
  }

  // Hide current reticule and update button UI.
  function fovButtonHide() {
    // Hide current reticule.
    fovMain[curRet].hideFov(aladin);

    // Make the current reticule background.
    fovMain[curRet].toBack();

    formMain[curRet].compDisplay = false;

    // Update the form UI.
    formMain[curRet].buttonFovForm(String.fromCodePoint(0xE8F4));
  }

  // *** Respond to FOV Show/Hide button.
  $('#fovShow').on('keyup click', function() {
    if (formMain[curRet].compDisplay) {
      // Reticule is displayed so hide it.
      fovButtonHide();
    } else {
      // Reticule is hidden so display it.
      fovButtonShow();
    }
  });

  // ******* Compare Form

  // *** Respond to changes in the Description field.
  $('#compDesc').on('change', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getCompForm();

    if (formMain[curRet].compDisplay) {
      // Show the FoV reticule.
      fovButtonShow();
    } else {
      // Hide the FoV reticule (only effect is to update FOV form UI).
      fovButtonHide();
    }

    // Update the form UI with the values used.
    formMain[curRet].putCompForm();
    formMain[curRet].dirtyUrlForm();
    compNavigate();
  });

  // *** Respond to change in Width input.
  $('#compWidth').on('change', function() {
    // Clear any computed pixel scale and scope resolution values.
    formMain[curRet].fovAppH = 0;
    formMain[curRet].fovAppV = 0;
    formMain[curRet].fovScopeRes = 0;

    // Show the FoV reticule.
    fovButtonShow();

    // Update the form UI.
    formMain[curRet].dirtyUrlForm();
  });

  // *** Respond to change in Height input.
  $('#compHeight').on('change', function() {
    // Clear any computed pixel scale and scope resolution values.
    formMain[curRet].fovAppH = 0;
    formMain[curRet].fovAppV = 0;
    formMain[curRet].fovScopeRes = 0;

    // Show the FoV reticule.
    fovButtonShow();

    // Update the form UI.
    formMain[curRet].dirtyUrlForm();
  });

  // *** Respond to Maximise checkbox.
  $('#compMax').on('change', function() {
    // Show the FoV reticule.
    fovButtonShow();

    // Update the form UI.
    formMain[curRet].dirtyUrlForm();
  });

  // *** Respond to Annotate checkbox.
  $('#compNotes').on('change', function() {
    // Show the FoV reticule.
    fovButtonShow();

    // Update the form UI.
    formMain[curRet].dirtyUrlForm();
  });

  // *** Respond to change in colour selection.
  $('#compColour').on('change', function() {
    // Show the FoV reticule.
    fovButtonShow();

    // Update the form UI.
    formMain[curRet].dirtyUrlForm();
  });

  // *** Respond to Compare New button.
  $('#compNew').on('keyup click', function() {
    // Make the current reticule background.
    fovMain[curRet].toBack();

    // Append new default reticule.
    curRet = fovMain.length;
    fovMain[curRet] = new fovReticule(2.13, 1.42, 0, 0, 0, '#00FF00', 0, 0, 0, true, true, 'New');
    formMain[curRet] = new itForm();
    aladin.addCanvas(fovMain[curRet].guid);

    // Grab target from previously displayed form.
    formMain[curRet].getSearchForm();
    if (!formMain[curRet].isBlankGotoForm()) {
      formMain[curRet].getGotoForm();
    }

    // Update the form UI with the values used.
    formMain[curRet].putAllForm();
    formMain[curRet].dirtyUrlForm();
    compNavigate();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to Compare Copy button.
  $('#compCopy').on('keyup click', function() {
    // Make the current reticule background.
    fovMain[curRet].toBack();

    // Append new default reticule.
    curRet = fovMain.length;
    fovMain[curRet] = new fovReticule(2.13, 1.42, 0, 0, 0, '#00FF00', 0, 0, 0, true, true, 'New');
    formMain[curRet] = new itForm();
    aladin.addCanvas(fovMain[curRet].guid);

    // Grab everything from previously displayed form.
    formMain[curRet].getSearchForm();
    if (!formMain[curRet].isBlankGotoForm()) {
      formMain[curRet].getGotoForm();
    }
    formMain[curRet].getFovForm();
    formMain[curRet].getCcdForm();
    formMain[curRet].getPixelForm();
    formMain[curRet].getEquipForm();
    formMain[curRet].getCompForm();

    formMain[curRet].compDesc = (formMain[curRet].compDesc.length > 15) ? formMain[curRet].compDesc.slice(0, 15) + '..Copy' : formMain[curRet].compDesc + '..Copy';

    // Update the form UI with the values used.
    formMain[curRet].putAllForm();
    formMain[curRet].dirtyUrlForm();
    compNavigate();

    // Show the FoV reticule.
    fovButtonShow();
  });

  // *** Respond to Compare Delete button.
  $('#compDelete').on('keyup click', function() {
    // Hide current reticule if displayed.
    if (formMain[curRet].compDisplay) {
      fovButtonHide();
    }

    // Delete current reticule.
    aladin.removeCanvas(fovMain[curRet].guid);
    formMain.splice(curRet, 1);
    fovMain.splice(curRet, 1);

    // Display another reticule.
    if (curRet > formMain.length - 1) {
      if (curRet === 0) {
        // Add a new default reticule.
        fovMain[curRet] = new fovReticule(2.13, 1.42, 0, 0, 0, '#00FF00', 0, 0, 0, true, true, 'New');
        formMain[curRet] = new itForm();
        aladin.addCanvas(fovMain[curRet].guid);

        // Grab target from previously displayed form.
        formMain[curRet].getSearchForm();
        if (!formMain[curRet].isBlankGotoForm()) {
          formMain[curRet].getGotoForm();
        }
      } else {
        // Beyond end of list, move back one.
        curRet = curRet - 1;
      }
    } // else still in list.

    // Update the form UI with the values used.
    formMain[curRet].putAllForm();
    formMain[curRet].dirtyUrlForm();
    compNavigate();

    if (formMain[curRet].compDisplay) {
      // Show the FoV reticule.
      fovButtonShow();
    } else {
      // Hide the FoV reticule (only effect is to update FOV form UI).
      fovButtonHide();
    }
  });

  // ******* Search Form

  // *** Respond to the object search button.
  $('#searchTargetGo').on('keyup click', function() {
    // Fetch the contents of the search form.
    formMain[curRet].getSearchForm();

    // Clear the other inputs relating to finding a target.
    formMain[curRet].searchPopular = '';
    formMain[curRet].clearGotoForm();

    // Update the UI.
    formMain[curRet].putSearchForm();
    formMain[curRet].putGotoForm();

    // Make Aladin Lite attempt to resolve and go to the named target.
    aladin.gotoObject(formMain[curRet].searchTarget);
  });

  // *** Respond to changes in the browse selection list.
  $('#searchPopular').on('change', function() {
    // Fetch the contents of the Search form.
    formMain[curRet].getSearchForm();

    // Clear the other inputs relating to finding a target.
    formMain[curRet].searchTarget = '';
    formMain[curRet].clearGotoForm();

    // Update the UI.
    formMain[curRet].putSearchForm();
    formMain[curRet].putGotoForm();

    // Make Aladin Lite attempt to resolve and go to the named target.
    aladin.gotoObject(formMain[curRet].searchPopular);
  });

  // *** Respond to changes in the survey selection list.
  $('#searchSurveySelection').on('change', function() {
    // Fetch the contents of the Search form.
    formMain[curRet].getSearchForm();

    // Update the UI.
    formMain[curRet].putSearchForm();

    // Make Aladin Lite change the displayed survey.
    aladin.setImageSurvey(formMain[curRet].searchSurveySelection);
  });

  // ******* RA/Dec Form

  // *** Respond to Goto RA / Dec button.
  $('#gotoRaDec').on('keyup click', function() {
    if (!formMain[curRet].isBlankGotoForm()) {
      // Fetch the contents of the Goto form.
      formMain[curRet].getGotoForm();

      // Clear the other inputs relating to finding a target.
      formMain[curRet].searchTarget = '';
      formMain[curRet].searchPopular = '';

      // Update the UI.
      formMain[curRet].putGotoForm();
      formMain[curRet].putSearchForm();

      // Make Aladin Lite go to the coordinates.
      aladin.gotoPosition(formMain[curRet].gotoRaHD, formMain[curRet].gotoDecDD);
    }
  });

  // *** Respond to Get RA / Dec button.
  $('#getRaDec').on('keyup click', function() {
    // Get RA and Dec from Aladin.
    formMain[curRet].gotoRaHD = aladin.view.viewCenter.lon;
    formMain[curRet].gotoDecDD = aladin.view.viewCenter.lat;

    // Convert retrieved decimal RA and Dec to sexagesimal.
    formMain[curRet].rdsGotoForm();
    formMain[curRet].ddsGotoForm();

    // Clear the other inputs relating to finding a target.
    formMain[curRet].searchTarget = '';
    formMain[curRet].searchPopular = '';

    // Update the UI.
    formMain[curRet].putGotoForm();
    formMain[curRet].putSearchForm();
  });

  // ******* URL Form

  // *** Respond to URL Create button.
  $('#urlCreate').on('keyup click', function() {
    // Get the current RA/Dec since that the only way to get back to current target.
    $('#getRaDec').click();

    // Create URL from form parameters and put it in link box.
    var newSave = formMain[curRet].createSavedForm(fovMain[curRet]);
    formMain[curRet].putUrlForm('http://www.blackwaterskies.co.uk/imaging-toolbox/?BWSFOV=' +
      encodeURIComponent(JSON.stringify(newSave)));

    // Select all in the link box to make copy/paste easier for the user.
    $('#urlSave').select();
  });

  // *** Respond to URL Save button.
  $('#urlStore').on('keyup click', function() {
    // Get the current RA/Dec since that the only way to get back to current target.
    $('#getRaDec').click();

    // Create an array of saved reticules.
    var reticuleSet = [];
    for (var i = 0, len = formMain.length; i < len; i++) {
      reticuleSet[i] = formMain[i].createSavedForm(fovMain[i]);
    }

    // Store the set of reticules in local storage.
    localStorage.setItem('reticuleSet', JSON.stringify(reticuleSet));

    // Save the current reticule index.
    localStorage.setItem('curRet', curRet);

    // Update the form UI.
    formMain[curRet].cleanUrlForm();
  });

  // *** Respond to URL PNG button.
  $('#urlPng').on('keyup click', function() {
    // Create a canvas to export the current view.
    var exportCanvas = document.createElement("canvas");
    exportCanvas.width = aladin.getSize()[0];
    exportCanvas.height = aladin.getSize()[1];

    // Draw all the existing canvases on the export canvas.
    var exportCtx = exportCanvas.getContext("2d");

    // Draw the static Aladin Lite canvases first.
    exportCtx.drawImage(aladin.view.imageCanvas, 0, 0);
    exportCtx.drawImage(aladin.view.catalogCanvas, 0, 0);
    exportCtx.drawImage(aladin.view.reticleCanvas, 0, 0);

    // Loop through all FOV reticule canvases and export them.
    for (var i = 0, len = fovMain.length; i < len; i++) {
      exportCtx.drawImage($('#' + fovMain[i].guid)[0], 0, 0);
    }

    // Export the canvas to a new window/tab as a PNG.
    var iframeWin = window.open(); // Necessary as Chrome no longer allows top level windows to be populated with DataURLs
    iframeWin.title = 'Imaging Toolbox Export';
    iframeWin.document.write('<iframe id="FOVExport" src=' + exportCanvas.toDataURL('image/png') + ' frameborder="0" style="border:0; top:0px; left:0px; bottom:0px; right:0px; width:' + exportCanvas.width + '; height:' + exportCanvas.height + ';" allowfullscreen></iframe>');
  });

  // ******* CCD Form

  // *** Respond to CCD Calculate button.
  $('#ccdCalculate').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getCcdForm();

    // Update other forms with required values.
    formMain[curRet].pixelFocalLength = formMain[curRet].ccdFocalLength;
    formMain[curRet].populateFovForm(formMain[curRet].ccdWidth, formMain[curRet].ccdHeight, formMain[curRet].ccdFocalLength, 0, 0, 0);

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].putCcdForm();
    formMain[curRet].putPixelForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV overlay.
    fovButtonShow();
  });

  // ******* Pixel Form

  // *** Respond to Pixel Calculate button.
  $('#pixelCalculate').on('keyup click', function() {
    // Get and sanitise parameters from the form UI.
    formMain[curRet].getPixelForm();

    // Calculate the chip size from the pixel size and resolution.
    formMain[curRet].ccdWidth = formMain[curRet].pixelWidth * formMain[curRet].pixelHRes / 1000;
    formMain[curRet].ccdHeight = formMain[curRet].pixelHeight * formMain[curRet].pixelVRes / 1000;

    // Update other forms with required values.
    formMain[curRet].ccdFocalLength = formMain[curRet].pixelFocalLength;
    formMain[curRet].populateFovForm(formMain[curRet].ccdWidth, formMain[curRet].ccdHeight, formMain[curRet].pixelFocalLength,
      formMain[curRet].pixelHRes, formMain[curRet].pixelVRes, formMain[curRet].pixelAperture);

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].putCcdForm();
    formMain[curRet].putPixelForm();
    formMain[curRet].dirtyUrlForm();

    // Show the FoV overlay.
    fovButtonShow();
  });

  // ******* Equipment Form

  // *** Respond to Equipment Calculate button.
  $('#equipCalculate').on('keyup click', function() {
    // Get and sanitise parameters from the form UI, including parsing of the camera and scope objects.
    formMain[curRet].getEquipForm();

    // Update the form UI with the values used.
    formMain[curRet].putFovForm();
    formMain[curRet].putCcdForm();
    formMain[curRet].putPixelForm();
    formMain[curRet].putEquipForm();
    formMain[curRet].dirtyUrlForm();

    // Use the Pixel FOV form to calculate everything else.
    $('#pixelCalculate').click();
  });
} // ****** Activation Section End.

// *** Update the Compare form UI based on current reticule selected.
function compNavigate() {
  // Update the Reticule form header.
  $('#fovForm div.fhead').text('Reticule (' + formMain[curRet].compDesc + ')');

  // Update the Compare form header.
  $('#reticuleCount').text('(' + (curRet + 1) + '/' + formMain.length + ')');

  // Update the Compare form navigation buttons.
  if (curRet == formMain.length - 1) {
    // End of list.
    $('#fovNext').prop('disabled', true);
  } else {
    $('#fovNext').prop('disabled', false);
  }
  if (curRet === 0) {
    // Start of list.
    $('#fovPrev').prop('disabled', true);
  } else {
    $('#fovPrev').prop('disabled', false);
  }
}

// ********** Form UI end *************************************************************************************
